#include <thread>
#include <vector>
#include <mutex>
#include <iostream>
#include <optional>

using namespace std;


vector<int> queue;     // Queue filled by producer and emptied by consumers.
mutex queueMutex;      // Mutex for queue.
bool finished = false;  // Flag indicating that all values were produced (but not necessarily consumed!).

void producer(int numValuesToProduce) {
    for (int i = 0; i < numValuesToProduce; i++) {
        // Value to be consumed.
        int value = std::rand();

        {
            lock_guard<mutex> queueLock(queueMutex);
            queue.insert(queue.begin(), value);
        }

        this_thread::sleep_for(chrono::milliseconds(rand() % 100));
    }

    {
        // Due to busy waiting, locking on queueMutex is not necessary.
//        lock_guard<mutex> queueLock(queueMutex);
        finished = true;
    }
}

void consumer(int id) {
    // Perform the loop until all the values are consumed.
    while (true) {
        {
            // Check if everything was consumed and no more values will be produced.
            lock_guard<mutex> queueLock(queueMutex);
            if (queue.empty() && finished) {
                return;
            }
        }

        optional<int> value = nullopt;

        {
            // Consume the value (if there is some).
            lock_guard<mutex> queueLock(queueMutex);
            if (!queue.empty()) {
                value = queue.back();
                queue.pop_back();
            }
        }
    }
}

int main() {
    int numConsumers = 4;  // Number of consumers.
    int numValuesToProduce = 100;   // Number of values to produce.

    vector<thread> workers;
    for (int consumerIndex = 0; consumerIndex < numConsumers; consumerIndex++) {
        workers.push_back(thread(consumer, consumerIndex));
    }

    workers.push_back(thread(producer, numValuesToProduce));

    for (auto &worker : workers) {
        worker.join();
    }

    cout << "All workers finished." << endl;
    cout << "Num remaining values queue: " << queue.size() << endl;    // Check that all values were consumed.
}
