#include <thread>
#include <vector>
#include <mutex>
#include <condition_variable>
#include <iostream>

using namespace std;


vector<int> queue;     // Queue filled by producer and emptied by consumers.
mutex queueMutex;      // Mutex for queue.
condition_variable queueCondVar;    // Condition variable for waiting/signalling on queue.
bool finished = false; // Flag indicating that all values were produced (but not necessarily consumed!).

void producer(int numValuesToProduce) {
    for (int i = 0; i < numValuesToProduce; i++) {
        // Value to be consumed.
        int value = std::rand();

        {
            lock_guard<mutex> queueLock(queueMutex);
            queue.insert(queue.begin(), value);
        }

        queueCondVar.notify_one();

        this_thread::sleep_for(chrono::milliseconds(rand() % 100));
    }

    {
        unique_lock<mutex> queueLock(queueMutex);
        finished = true;
    }
    queueCondVar.notify_all();
}

void consumer(int id) {
    while (true) {
        {
            // Wait on queueCondVar.
            unique_lock<mutex> queueLock(queueMutex);
            queueCondVar.wait(queueLock, [&]() {
                // 'queueMutex' is locked here.
                return finished || !queue.empty();
            });

            // 'queueMutex' is locked here.
            if (queue.empty() && finished) {
                return;
            }

            // Queue is not empty since we are not finished and the lock on queueLock in this was not released since
            // the predicate check
            int value = queue.back();
            queue.pop_back();
        }
    }
}

int main() {
    int numConsumers = 4;  // Number of consumers.
    int numValuesToProduce = 100;   // Number of values to produce.

    vector<thread> workers;
    for (int consumerIndex = 0; consumerIndex < numConsumers; consumerIndex++) {
        workers.push_back(thread(consumer, consumerIndex));
    }

    workers.push_back(thread(producer, numValuesToProduce));

    for (auto &worker : workers) {
        worker.join();
    }

    cout << "All workers finished." << endl;
    cout << "Num remaining values queue: " << queue.size() << endl;    // Check that all values were consumed.
}
